proc install {database_handle package_stream package} {

	# Insert uninstall proc into database, it will be called in
	# case of package partial installation due to unsolvable problems
	$package uninstall_proc [uninstall_proc]
	$package name [package_name]

	set err [check_features $database_handle $package_stream]
	if {[string length $err] > 0} {return $err}

	set err [check_installation_over_installed $database_handle $package_stream]
	if {[string length $err] > 0} {return $err}

	set err [check_reverse_conflicts $database_handle $package_stream]
	if {[string length $err] > 0} {return $err}

	set package_structure [structure]
	set err [check_files_installation $database_handle $package_stream [$package_structure files]]
	if {[llength $err] > 0} {
		return $err
	}

	# Install files supplied with the package
	set err [install_files $database_handle $package_stream $package [$package_structure files]]
	if {[string length $err] > 0} {
		return $err
	}

	# Update database
	$package description [package_description]
	$package features [provides]
	$package requires [requires]
	$package conflicts [conflicts]
	$package could_be_uninstalled_proc [could_be_uninstalled_proc]
	$package could_be_installed_proc [could_be_installed_proc]

	# Install subpackages
	if [$package_structure has_subpackages] {
		foreach subpackage [Callback::install_subpackages $package_structure] {
	                set err [check_files_installation $database_handle $package_stream [$subpackage files]]
	                if {[llength $err] == 0} {
				set err [install_files $database_handle $package_stream $package [$subpackage files]]
				if {[string length $err] > 0} {
					break
				}
	                } else {
				break
			}
		}
	}

	if {[string length $err] > 0} {
	        # Install references
	        if [$package_structure has_references] {
	        }
	}

	return $err
}

proc check_features {database_handle package_stream} {

	# Check if there are requirements, conflicts, users problems.
	# If promlems exist and not solved by callback returns error message,
	# otherwise returns an empty string

	while 1 {
		set requirements_problems [check_requirements $database_handle]
		set conflicts_problems [check_conflicts $database_handle]
		set users_problems [check_users_requirements]
		set groups_problems [check_groups_requirements]
		set suggestions_dont_met [check_suggestions $database_handle]
		if {[llength $requirements_problems] == 0 && [llength $conflicts_problems] == 0 && [llength $users_problems] == 0 && [llength $groups_problems] == 0 && [llength $suggestions_dont_met] == 0} {
			break
		}
	        set callback_result [Callback::package_requirements_not_met $database_handle $package_stream $requirements_problems $conflicts_problems $users_problems $groups_problems $suggestions_dont_met]
	        switch $callback_result {
	                "" { continue }
	                "ignore" { break }
	                default { return $callback_result }
	        }
	}
	return ""
}

proc check_installation_over_installed {database_handle package_stream} {

	# Check if features provided by the package are already provided by the system.
	# returns error message or an empty string if there are no problems.

	while 1 {
		set features_of_package [provides]
		set duplicated_features {}
		foreach feature_provided $features_of_package {
		        foreach sys_feature [$database_handle feature [$feature_provided name] provide] {
		                if {[$sys_feature equal $feature_provided]} {
		                        lappend duplicated_features $sys_feature
		                }
		        }
		}
	        set callback_called 0
		if {[llength $duplicated_features] > 0} {
		        set callback_result [Callback::installation_over_installed $database_handle $package_stream $duplicated_features]
		        set callback_called 1
		}
		if $callback_called {
			if {[string length $callback_result] > 0} {
				return $callback_result
			}
		} else {
			break
		}
	}
	return ""
}

proc check_reverse_conflicts {database_handle package_stream} {

	# Check if packages installed have features provided by the package
        # being installed in their conflicting lists.
	# returns error message or an empty string if there are no problems.

	while 1 {
		set conflicting_packages {}
		set features_of_package [provides]
		foreach feature_provided $features_of_package {
			set conflicting_packages [concat $conflicting_packages [$database_handle packages_conflicting $feature_provided]]
		}
	        set callback_called 0
		if {[llength $conflicting_packages] >0} {
		        set callback_result [Callback::reverse_conflict $database_handle $package_stream $conflicting_packages]
		        set callback_called 1
		}
		if $callback_called {
			if {[string length $callback_result] > 0} {
				return $callback_result
			}
		} else {
			break
		}
	}
	return 
}

proc check_files_installation {database_handle package_stream files} {

	# Check installation problems for the files supplied with the package

	set err {}
	foreach file $files {
		if [$package_stream file_exists $file] {
			set owners [$database_handle packages_owning [$package_stream file_pathname_for_installation $file]]
			if {[$file attached] == "file"} {
				switch [$file replacement_scheme] {
					"always_replace" {}
					"never_replace" {}
					default {
						if {[llength $owners] > 0} {
							set err_ret [Callback::file_to_be_installed_exists_and_owned_by_other_packages $database_handle $package_stream $file $owners]
							if {[string length $err_ret] > 0} {
								lappend err $err_ret
							}
						}
					}
				}
			}
		} else {
			if {[$file attached] == "patch"} {
				lappend err [format "File %s is to be patched but does not exist" [$file pathname]]
				break
			}
		}
	}
	return $err
}

proc install_files {database_handle package_stream package files} {
	set error {}
	foreach file $files {
		if [$package_stream file_exists $file] {
			set owners [$database_handle packages_owning [$package_stream file_pathname_for_installation $file]]
			if {[$file attached] == "file"} {
				switch [$file replacement_scheme] {
					"always_replace" {set installed_file [$package_stream install_file $file]}
					"never_replace" {set installed_file [$package_stream file_pathname_for_installation $file]}
					default {set installed_file [$package_stream install_file $file]}
				}
			} elseif {[$file attached] == "patch"} {
				set installed_file [$package_stream patch_file $file]
			} elseif {[$file attached] == "append"} {
				set installed_file [$package_stream append_file $file]
			} else {
				if {[string length [$file symlink]] > 0} {
					set installed_file [$package_stream create_symlink $file]
				} elseif {[llength [$file device]] > 0} {
					set installed_file [$package_stream create_device $file]
				}
			}
		} else {
			if {[$file attached] == "patch"} {
				set error [format "File %s is to be patched but does not exist" [$file pathname]]
			} elseif {[$file attached] == "file"} {
				set installed_file [$package_stream install_file $file]
			} elseif {[$file attached] == "append"} {
				set installed_file [$package_stream append_file $file create]
			} else {
				if {[string length [$file symlink]] > 0} {
					set installed_file [$package_stream create_symlink $file]
				} elseif {[llength [$file device]] > 0} {
					set installed_file [$package_stream create_device $file]
				}
			}
		}
		if {[string length $error] == 0} {
			$package own $installed_file
		}
		if {[string length $error] > 0} {
			break
		}
	}
				return $error
}


proc upgrade {database_handle package_stream package} {

	install $database_handle $package_stream $package
}

proc uninstall_proc {} { return {
proc uninstall {database_handle package} {

	# Check if other packages require presence of this package
	set packages_requiring [$database_handle packages_requiring [$package features]]
	set packages_really_requiring {}
	foreach package_requiring $packages_requiring {
		if {[$package_requiring could_be_uninstalled [$package features]] == 0} {
			lappend packages_really_requiring $package_requiring
		}
	}
	if {[llength $packages_really_requiring] > 0} {
		set callback_result [Callback::uninstall_required_features $database_handle $package $packages_really_requiring]
		if {[string length $callback_result] > 0} {
			return $callback_result
		}
	}

	# Uninstall owned files
	foreach file [$package files] {
		$package uninstall_file $file
	}

	# Erase package database record
	$package remove

	return ""
}
}
}
